package org.eclipse.jdt.internal.compiler.lookup;

package .var.folders._j.vfbv187d0bb8b8w5spzgg8180000gn.T;

import org.eclipse.jdt.internal.compiler.lookup.InferenceContext18;
import org.eclipse.jdt.internal.compiler.lookup.InferenceVariable;
import org.eclipse.jdt.internal.compiler.lookup.InvocationSite;
import org.eclipse.jdt.internal.compiler.lookup.LookupEnvironment;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.Substitution;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeVariableBinding;

public class InferenceSubstitution extends Scope.Substitutor implements Substitution {
  private LookupEnvironment environment;
  
  private InferenceVariable[] variables;
  
  private InvocationSite[] sites;
  
  public InferenceSubstitution(LookupEnvironment environment, InferenceVariable[] variables, InvocationSite site) {
    this.environment = environment;
    this.variables = variables;
    this.sites = new InvocationSite[] { site };
  }
  
  public InferenceSubstitution(InferenceContext18 context) {
    this(context.environment, context.inferenceVariables, context.currentInvocation);
  }
  
  public org.eclipse.jdt.internal.compiler.lookup.InferenceSubstitution addContext(InferenceContext18 otherContext) {
    Object object = new Object(this, this.environment, null, null);
    int l1 = this.sites.length;
    ((org.eclipse.jdt.internal.compiler.lookup.InferenceSubstitution)object).sites = new InvocationSite[l1 + 1];
    System.arraycopy(this.sites, 0, ((org.eclipse.jdt.internal.compiler.lookup.InferenceSubstitution)object).sites, 0, l1);
    ((org.eclipse.jdt.internal.compiler.lookup.InferenceSubstitution)object).sites[l1] = otherContext.currentInvocation;
    ((org.eclipse.jdt.internal.compiler.lookup.InferenceSubstitution)object).variables = this.variables;
    return (org.eclipse.jdt.internal.compiler.lookup.InferenceSubstitution)object;
  }
  
  public TypeBinding substitute(Substitution substitution, TypeBinding originalType) {
    for (int i = 0; i < this.variables.length; i++) {
      InferenceVariable variable = this.variables[i];
      if (isInSites(variable.site) && isSameParameter(getP(i), originalType)) {
        if (this.environment.globalOptions.isAnnotationBasedNullAnalysisEnabled && originalType.hasNullTypeAnnotations())
          return this.environment.createAnnotatedType(variable.withoutToplevelNullAnnotation(), originalType.getTypeAnnotations()); 
        return (TypeBinding)variable;
      } 
    } 
    return super.substitute(substitution, originalType);
  }
  
  private boolean isInSites(InvocationSite otherSite) {
    for (int i = 0; i < this.sites.length; i++) {
      if (InferenceContext18.isSameSite(this.sites[i], otherSite))
        return true; 
    } 
    return false;
  }
  
  protected boolean isSameParameter(TypeBinding p1, TypeBinding originalType) {
    return TypeBinding.equalsEquals(p1, originalType);
  }
  
  protected TypeBinding getP(int i) {
    return (this.variables[i]).typeParameter;
  }
  
  public TypeBinding substitute(TypeVariableBinding typeVariable) {
    InferenceVariable inferenceVariable;
    ReferenceBinding superclass = typeVariable.superclass;
    ReferenceBinding[] superInterfaces = typeVariable.superInterfaces;
    boolean hasSubstituted = false;
    for (int i = 0; i < this.variables.length; i++) {
      InferenceVariable variable = this.variables[i];
      TypeBinding pi = getP(i);
      if (TypeBinding.equalsEquals(pi, (TypeBinding)typeVariable))
        return (TypeBinding)variable; 
      if (TypeBinding.equalsEquals(pi, (TypeBinding)superclass)) {
        inferenceVariable = variable;
        hasSubstituted = true;
      } else if (superInterfaces != null) {
        int ifcLen = superInterfaces.length;
        for (int j = 0; j < ifcLen; j++) {
          if (TypeBinding.equalsEquals(pi, (TypeBinding)superInterfaces[j])) {
            if (superInterfaces == typeVariable.superInterfaces)
              System.arraycopy(superInterfaces, 0, superInterfaces = new ReferenceBinding[ifcLen], 0, ifcLen); 
            superInterfaces[j] = (ReferenceBinding)variable;
            hasSubstituted = true;
            break;
          } 
        } 
      } 
    } 
    if (hasSubstituted) {
      typeVariable = new TypeVariableBinding(typeVariable.sourceName, typeVariable.declaringElement, typeVariable.rank, this.environment);
      typeVariable.superclass = (ReferenceBinding)inferenceVariable;
      typeVariable.superInterfaces = superInterfaces;
      typeVariable.firstBound = (inferenceVariable != null) ? (TypeBinding)inferenceVariable : (TypeBinding)superInterfaces[0];
      if (typeVariable.firstBound.hasNullTypeAnnotations())
        typeVariable.tagBits |= 0x100000L; 
    } 
    return (TypeBinding)typeVariable;
  }
  
  public LookupEnvironment environment() {
    return this.environment;
  }
  
  public boolean isRawSubstitution() {
    return false;
  }
}
